/* $Id: tools.h,v 2.2.5 1996/03/25 19:17:43 root Exp $
 *
 * ISDN accounting for isdn4linux.
 *
 * Copyright 1995, 1996 by Andreas Kool (akool@Kool.f.EUnet.de)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Log: tools.h,v $
 * Revision 2.2.5  1996/03/25  19:17:43  akool
 * 1TR6 causes implemented
 *
 * Revision 2.23  1996/03/24  12:11:43  akool
 * Explicit decl. of basename() - new "unistd.h" dont have one
 *
 * Revision 2.15  1996/02/21  20:14:43  akool
 *
 * Revision 2.12  1996/02/13  20:08:43  root
 * Nu geht's (oder?)
 *
 * Revision 2.12  1996/02/13  20:08:43  root
 * Nu geht's (oder?)
 *
 * Revision 1.2  1996/02/13  20:05:28  root
 * so nun gehts
 *
 * Revision 1.1  1996/02/13  14:28:14  root
 * Initial revision
 *
 * Revision 2.05  1995/02/11  17:10:16  akool
 *
 */

/****************************************************************************/

#ifndef _TOOLS_H_
#define _TOOLS_H_

/****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <malloc.h>
#include <time.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <fcntl.h>
#include <ctype.h>
#include <limits.h>
#include <signal.h>
#include <math.h>
#include <syslog.h>
#ifdef linux
#include <ndbm.h>
#include <sys/kd.h>
extern char *basename __P((__const char *__name));
#else
#include "/usr/ucbinclude/ndbm.h"
#include <libgen.h>
#endif

/****************************************************************************/

#undef  min
#undef  max
#define max(a,b)        (((a) > (b)) ? (a) : (b))
#define min(a,b)        (((a) < (b)) ? (a) : (b))

/****************************************************************************/

#define NIL  (char **)NULL

/****************************************************************************/

#define NUMSIZE      20
#define RETSIZE      32
#define MAXRET	      5
#define MAXZONES      6
#define MAXCHAN       6

#define DIGITS 	     17
#define DEB           1

#define MAXUNKNOWN   50
#define MAXCONNECTS  50

/****************************************************************************/

#define CALLING       0
#define CALLED        1
#define DATETIME      2

/****************************************************************************/

#define DIALOUT	      0
#define DIALIN	      1

/****************************************************************************/

#define ALERTING               0x01
#define CALL_PROCEEDING	       0x02
#define SETUP                  0x05
#define SETUP_ACKNOWLEDGE      0x0d
#define SUSPEND                0x25
#define SUSPEND_ACKNOWLEDGE    0x2d
#define RESUME                 0x26
#define RESUME_ACKNOWLEDGE     0x2e
#define CONNECT                0x07
#define CONNECT_ACKNOWLEDGE    0x0f
#define FACILITY               0x62
#define NOTIFY                 0x6e
#define STATUS                 0x7d
#define MAKEL_ACKNOWLEDGE      0x28
#define MAKEL_RESUME_ACK       0x33
#define DISCONNECT             0x45
#define RELEASE                0x4d
#define RELEASE_COMPLETE       0x5a
#define AOCD_1TR6              0x6d

/****************************************************************************/

#define	RING_INCOMING	          1
#define RING_OUTGOING		  2
#define RING_CONNECT		  4
#define	RING_RING		  8
#define	RING_KILL		 16
#define RING_PROVIDER		 32
#define RING_AOCD		 64
#define RING_SPEAK		128

/****************************************************************************/

#define STATE_RING		  1 /* "Telefonklingeln" ... jemand ruft an, oder man selbst ruft raus */
#define STATE_CONNECT		  2 /* Verbindung */
#define STATE_HANGUP		  3 /* Verbindung beendet */
#define STATE_AOCD		100 /* Gebuehrenimpuls _waehrend_ der Verbindung */
#define	STATE_CAUSE		101 /* Aussergewoehnliche Cause-Meldungen von der VSt */
#define STATE_TIME		102 /* Uhrzeit-Meldung von der VSt */
#define STATE_BYTE		103 /* Durchsatz-Meldung von Frank (Byte/s/B-Kanal) */

/****************************************************************************/

#define AOC_130	           -999990L

/****************************************************************************/

#define QCMASK  	       0377
#define QUOTE   	       0200
#define QMASK      (QCMASK &~QUOTE)
#define NOT                     '!'

#define	AVON	             "avon"
#define INFO	    "/dev/isdninfo"

#define	BIGBUFSIZ              2048

/****************************************************************************/

/* from "isdn.h" */
#define ISDN_USAGE_NONE       0
#define ISDN_USAGE_RAW        1
#define ISDN_USAGE_MODEM      2
#define ISDN_USAGE_NET        3
#define ISDN_USAGE_VOICE      4
#define ISDN_USAGE_FAX        5
#define ISDN_USAGE_MASK       7 /* Mask to get plain usage */
#define ISDN_USAGE_EXCLUSIVE 64 /* This bit is set, if channel is exclusive */
#define ISDN_USAGE_OUTGOING 128 /* This bit is set, if channel is outgoing  */

/****************************************************************************/

#define VERSION_UNKNOWN       0
#define	VERSION_EDSS1	      1
#define	VERSION_1TR6	      2

/****************************************************************************/

#define  OTHER (call[chan].dialin ? CALLING : CALLED)
#define  ME    (call[chan].dialin ? CALLED : CALLING)

/****************************************************************************/

#define SHORT_STRING_SIZE      256
#define LONG_STRING_SIZE      1024
#define BUF_SIZE              4096

/****************************************************************************/

typedef struct {
  int	  state;
  int     cref;
  int     tei;
  int     ocref;
  int	  dialin;
  int	  cause;
  int	  aoce;
  int	  net;
  int	  self;
  int	  self2;
  int	  dialog;
  int	  release;
  int     bearer;
  int	  sock;
  char    num[2][NUMSIZE];
  char    vnum[2][256];
  char    id[32];
  char	  usage[16];
  int	  confentry[2];
  time_t  time;
  time_t  connect;
  time_t  localconnect;
  time_t  t_duration;
  time_t  disconnect;
  clock_t duration;
  long	  ibytes;
  long	  obytes;
  double  ibps;
  double  obps;
  char	  vorwahl[2][NUMSIZE];
  char	  rufnummer[2][NUMSIZE];
  char	  alias[2][NUMSIZE];
  char	  area[2][128];
  char	  money[64];
  char	  currency[32];
  char    msg[128];
  int     stat;
  int	  version;
} CALL;

/****************************************************************************/

typedef struct {
  char   *num;
  char	 *who;
  int	  zone;
  int	  flags;
  char	 *time;
  char   *infoarg;
  /* above from "isdnlog.conf" */
  int	  usage[2];
  double  dur[2];
  int     eh;
  double  dm;
} KNOWN;

/****************************************************************************/

typedef struct {
  int    in;
  int    out;
  int    eh;
  int    err;
  double din;
  double dout;
  double dm;
  long	 ibytes;
  long	 obytes;
} sum_calls;

/****************************************************************************/

typedef struct {
  int    eh;
  int    cause;
  time_t t;
  int    dir;
  double duration;
  double dm;
  char   num[2][NUMSIZE];
  long	 ibytes;
  long	 obytes;
} one_call;

/****************************************************************************/

PUBLIC KNOWN **known;
PUBLIC int     mymsns;
PUBLIC char   *myprefix;
PUBLIC int     knowns;
PUBLIC char   *hangup;
PUBLIC char   *aocd;
PUBLIC double  currency_factor;
PUBLIC char   *currency;
PUBLIC int     day;
PUBLIC int     month;
PUBLIC int     retnum;
PUBLIC int     ln;
PUBLIC char    retstr[MAXRET][RETSIZE];
PUBLIC char   *myname, *myshortname;
PUBLIC char   *Months[];
PUBLIC char   *CauseValue[];
PUBLIC datum   key, data;
PUBLIC DBM    *dbm;
PUBLIC time_t  cur_time;
PUBLIC double  einheit;

/****************************************************************************/

extern int   optind, errno;
extern char *optarg;

/****************************************************************************/

#ifdef _TOOLS_C_
#define _EXTERN
#else
#define _EXTERN extern
#endif

_EXTERN time_t atom(register char *p);
_EXTERN void   readconfig(char *myname);
_EXTERN void   discardconfig(void);
_EXTERN int    match(register char *s, register char *p);
_EXTERN char  *num2nam(char *num);
_EXTERN char  *double2str(double n, int l, int d, int flags);
_EXTERN char  *double2clock(double n);
_EXTERN char  *vnum(char *num, char *vorwahl, char *rufnummer, char *alias, char *area, int *confentry);
_EXTERN char  *i2a(int n, int l, int base);
_EXTERN char  *tr6cause(int cause);

#undef _EXTERN

#endif /* _TOOLS_H_ */
